<?php

declare(strict_types=1);

namespace Drupal\decoupled_passkeys\Services;

use Cose\Algorithm\Manager;
use Cose\Algorithm\Signature\ECDSA\ES256;
use Cose\Algorithm\Signature\RSA\RS256;
use Cose\Algorithms;
use Drupal\Component\Datetime\TimeInterface;
use Drupal\Core\Entity\EntityStorageInterface;
use Drupal\Core\Entity\EntityTypeManagerInterface;
use Drupal\Core\Logger\LoggerChannelTrait;
use Drupal\Core\TempStore\PrivateTempStore;
use Drupal\Core\TempStore\PrivateTempStoreFactory;
use Drupal\decoupled_passkeys\Exception\DecoupledPasskeysException;
use Drupal\decoupled_passkeys\Exception\DecoupledPasskeysInvalidArgumentException;
use Drupal\decoupled_passkeys\Exception\DecoupledPasskeysUnexpectedValueException;
use Drupal\decoupled_passkeys\Exception\DecoupledPasskeysValidationFailedException;
use Drupal\public_key_credential_source\Entity\PublicKeyCredentialSource;
use Drupal\user\UserInterface;
use Psr\Log\LoggerInterface;
use Symfony\Component\HttpFoundation\Request;
use Symfony\Component\HttpFoundation\RequestStack;
use Webauthn\AttestationStatement\AttestationObjectLoader;
use Webauthn\AttestationStatement\AttestationStatementSupportManager;
use Webauthn\AttestationStatement\NoneAttestationStatementSupport;
use Webauthn\AuthenticationExtensions\ExtensionOutputCheckerHandler;
use Webauthn\AuthenticatorAssertionResponse;
use Webauthn\AuthenticatorAssertionResponseValidator;
use Webauthn\AuthenticatorAttestationResponse;
use Webauthn\AuthenticatorAttestationResponseValidator;
use Webauthn\AuthenticatorSelectionCriteria;
use Webauthn\PublicKeyCredential as PKCredential;
use Webauthn\PublicKeyCredentialCreationOptions;
use Webauthn\PublicKeyCredentialDescriptor;
use Webauthn\PublicKeyCredentialLoader;
use Webauthn\PublicKeyCredentialParameters;
use Webauthn\PublicKeyCredentialRequestOptions;
use Webauthn\PublicKeyCredentialRpEntity;
use Webauthn\PublicKeyCredentialSource as PKCredentialSource;
use Webauthn\PublicKeyCredentialUserEntity;
use Webmozart\Assert\Assert;

/**
 * Helper class to facilitate use of webauthn-lib.
 *
 * Based on the docs:
 * https://webauthn-doc.spomky-labs.com/pure-php/the-hard-way.
 */
class WebauthnHelper {
  use LoggerChannelTrait;

  /**
   * The public key credential loader.
   *
   * @var \Webauthn\PublicKeyCredentialLoader
   */
  private PublicKeyCredentialLoader $publicKeyCredentialLoader;

  /**
   * The attestation response validator.
   *
   * @var \Webauthn\AuthenticatorAttestationResponseValidator
   */
  private AuthenticatorAttestationResponseValidator $authenticatorAttestationResponseValidator;

  /**
   * The assertion response validator.
   *
   * @var \Webauthn\AuthenticatorAssertionResponseValidator
   */
  private AuthenticatorAssertionResponseValidator $authenticatorAssertionResponseValidator;

  /**
   * The entity type manager.
   *
   * @var \Drupal\Core\Entity\EntityTypeManagerInterface
   */
  private EntityTypeManagerInterface $entityTypeManager;

  /**
   * The private temp store.
   *
   * @var \Drupal\Core\TempStore\PrivateTempStore
   */
  private PrivateTempStore $tempstore;

  /**
   * The logger.
   *
   * @var \Psr\Log\LoggerInterface
   */
  private LoggerInterface $logger;

  /**
   * The request stack.
   *
   * @var \Symfony\Component\HttpFoundation\RequestStack
   */
  private RequestStack $requestStack;

  /**
   * The entity storage interface for public_key_credential_source entities.
   *
   * @var \Drupal\Core\Entity\EntityStorageInterface
   */
  private EntityStorageInterface $publicKeyCredentialSourceEntityStorage;

  /**
   * The time interface.
   *
   * @var \Drupal\Component\Datetime\TimeInterface
   */
  private TimeInterface $time;

  /**
   * The key for storing PublicKeyCredentialCreationOptions.
   */
  private const TEMP_STORE_CREATION_OPTIONS = 'credential_creation_options';

  /**
   * The key for storing PublicKeyCredentialRequestOptions.
   */
  private const TEMP_STORE_REQUEST_OPTIONS = 'credential_request_options';

  /**
   * The name of the app for the public credential.
   *
   * @var string
   */
  private string $rpAppName = 'Drupal';

  /**
   * The icon encoded in base64 for the public credential.
   *
   * @var null
   */
  private null $rpIcon = NULL;

  /**
   * Webauthn helper constructor.
   */
  public function __construct(
    EntityTypeManagerInterface $entity_type_manager,
    PrivateTempStoreFactory $tempstore,
    RequestStack $request_stack,
    TimeInterface $time,
  ) {
    $attestation_statement_support_manager = AttestationStatementSupportManager::create();
    $attestation_statement_support_manager->add(NoneAttestationStatementSupport::create());

    $attestation_object_loader = AttestationObjectLoader::create($attestation_statement_support_manager);

    $this->publicKeyCredentialLoader = PublicKeyCredentialLoader::create(
      $attestation_object_loader
    );

    $extension_output_checker_handler = ExtensionOutputCheckerHandler::create();

    $algorithm_manager = Manager::create()
      ->add(
        ES256::create(),
        RS256::create()
      );

    $this->authenticatorAttestationResponseValidator = AuthenticatorAttestationResponseValidator::create(
      $attestation_statement_support_manager,
      // Deprecated Public Key Credential Source Repository. Please set null.
      NULL,
      // Deprecated Token Binding Handler. Please set null.
      NULL,
      $extension_output_checker_handler
    );

    $authenticatorAssertionResponseValidator = AuthenticatorAssertionResponseValidator::create(
    // Deprecated Public Key Credential Source Repository. Please set null.
      NULL,
      // Deprecated Token Binding Handler. Please set null.
      NULL,
      // The extension output checker handler.
      $extension_output_checker_handler,
      // The COSE Algorithm Manager.
      $algorithm_manager
    );
    $this->authenticatorAssertionResponseValidator = $authenticatorAssertionResponseValidator;

    $this->entityTypeManager = $entity_type_manager;
    $this->tempstore = $tempstore->get('decoupled_passkeys');
    $this->logger = $this->getLogger('decoupled_passkeys');
    $this->requestStack = $request_stack;
    $this->publicKeyCredentialSourceEntityStorage = $this->entityTypeManager->getStorage('public_key_credential_source');
    $this->time = $time;
  }

  /**
   * Gets the domain of the Drupal site request.
   *
   * @return string
   *   The domain of the Drupal site request.
   */
  private function getDrupalDomain(): string {
    $current_request = $this->requestStack->getCurrentRequest();
    if ($current_request instanceof Request) {
      return $current_request->getHost();
    }
    throw new DecoupledPasskeysInvalidArgumentException('Current request is not an HTTP request!');
  }

  /**
   * Generates an RP entity for the application (the Drupal site).
   */
  private function createRpEntity(): PublicKeyCredentialRpEntity {
    return PublicKeyCredentialRpEntity::create($this->rpAppName, $this->getDrupalDomain(), $this->rpIcon);
  }

  /**
   * Generates a webauthn user entity for the given user.
   *
   * @param \Drupal\user\UserInterface $user
   *   The Drupal user account.
   *
   * @return \Webauthn\PublicKeyCredentialUserEntity
   *   The public key credential entity for webauthn.
   */
  private function createUserEntity(UserInterface $user): PublicKeyCredentialUserEntity {
    $user_uuid = $user->uuid();
    $timestamp = $this->time->getCurrentTime();
    // @todo Sites may want to configure this.
    // Currently we set a timestamp for the username.
    // This is because the site is assumed to be using email registration.
    // If using usernames, it should be the Drupal username.
    $webauthn_user_name = gmdate('Y-m-d', $timestamp);
    $webauthn_user_id = $user_uuid;
    Assert::stringNotEmpty($webauthn_user_id, 'Failed to get UUID for user!');
    // @todo Sites may want to configure this.
    $webauthn_display_name = $user->getEmail();
    Assert::stringNotEmpty($webauthn_display_name, 'Failed to get user email for display name!');
    $webauthn_icon = NULL;
    return new PublicKeyCredentialUserEntity("$webauthn_user_name", $webauthn_user_id, $webauthn_display_name, $webauthn_icon);
  }

  /**
   * Generates a challenge to prevent man-in-the-middle attacks.
   *
   * @return string
   *   The challenge for public key credential confirmation.
   */
  private function createChallenge(): string {
    return random_bytes(16);
  }

  /**
   * Gets all credentials for a user.
   *
   * @param int $uid
   *   The user id of the user to get credentials for.
   *
   * @return array
   *   An array of webauthn-lib PublicKeyCredentialSources, not Drupal entities!
   */
  private function getAllPublicKeyCredentialSources(int $uid): array {
    $query = $this->publicKeyCredentialSourceEntityStorage->getQuery()->accessCheck(TRUE);
    $results = $query->condition('uid', $uid)
      ->execute();

    // Stop processing if there are no results.
    if (!is_array($results)) {
      return [];
    }

    /** @var \Drupal\public_key_credential_source\Entity\PublicKeyCredentialSource[] $all_public_key_credential_source_entities */
    $all_public_key_credential_source_entities = $this->publicKeyCredentialSourceEntityStorage->loadMultiple($results);
    $all_webauthn_credential_sources = array_map(static function (PublicKeyCredentialSource $entity) {
      return $entity->getPublicKeyCredentialSource();
    }, $all_public_key_credential_source_entities);

    return $all_webauthn_credential_sources;
  }

  /**
   * Gets the allowed credentials. Used for authentication and attestation.
   *
   * These are the excludeCredentials for registration,
   * and the valid credentials for authentication.
   *
   * @param string $user_email
   *   The email address of the user to get credentials for.
   *
   * @return array
   *   An array of Webauthn PublicKeyCredentialDescriptor objects.
   */
  public function getAllowedCredentials(string $user_email): array {
    $user = $this->getUserFromEmail($user_email);
    if (!($user instanceof UserInterface)) {
      throw new DecoupledPasskeysInvalidArgumentException("User email not associated with a Drupal account! $user_email");
    }
    $uid = intval($user->id());
    $all_drupal_credentials = $this->getAllPublicKeyCredentialSources($uid);

    // We don’t need the Credential Sources, just the associated Descriptors.
    $allowed_credentials = array_map(
      static function (PKCredentialSource $credential): PublicKeyCredentialDescriptor {
        return $credential->getPublicKeyCredentialDescriptor();
      },
      $all_drupal_credentials
    );
    // This value will be converted to JSON,
    // so the array keys need to be sequential.
    $reindexed_allowed_credentials = array_values($allowed_credentials);
    return $reindexed_allowed_credentials;
  }

  /**
   * Get Server side Registration options.
   *
   * @param \Drupal\user\UserInterface $user
   *   The registration target user.
   *
   * @return \Webauthn\PublicKeyCredentialCreationOptions
   *   Server side registration options.
   */
  public function getRegistrationOptions(UserInterface $user): PublicKeyCredentialCreationOptions {
    $user_entity = $this->createUserEntity($user);

    $rp_entity = $this->createRpEntity();

    $challenge = $this->createChallenge();

    // Defaults to ES256 + RS256 by default.
    // https://stackoverflow.com/a/77468132
    $public_key_credential_parameters_list = [
      PublicKeyCredentialParameters::create('public-key', Algorithms::COSE_ALGORITHM_ES256),
      PublicKeyCredentialParameters::create('public-key', Algorithms::COSE_ALGORITHM_RS256),
    ];

    // Even though these are default values, they must be set here
    // in order to return them in the JSON response.
    $authenticator_selection_criteria = AuthenticatorSelectionCriteria::create(
      AuthenticatorSelectionCriteria::AUTHENTICATOR_ATTACHMENT_NO_PREFERENCE,
      AuthenticatorSelectionCriteria::USER_VERIFICATION_REQUIREMENT_PREFERRED,
      AuthenticatorSelectionCriteria::RESIDENT_KEY_REQUIREMENT_PREFERRED,
      FALSE,
    );

    $user_email = $user->getEmail();
    if (!is_string($user_email)) {
      throw new DecoupledPasskeysUnexpectedValueException('Failed to get user uuid!');
    }
    $exclude_credentials = $this->getAllowedCredentials($user_email);

    $public_key_credential_creation_options = PublicKeyCredentialCreationOptions::create(
      $rp_entity,
      $user_entity,
      $challenge,
      $public_key_credential_parameters_list,
      $authenticator_selection_criteria,
      NULL,
      $exclude_credentials,
    );
    $this->tempstore->set($this::TEMP_STORE_CREATION_OPTIONS, $public_key_credential_creation_options);

    return $public_key_credential_creation_options;
  }

  /**
   * Gets the options for a user to authenticate (log in).
   *
   * @param string $user_email
   *   The email address of the user to authenticate.
   *
   * @return string
   *   The json stringified PublicKeyCredentialRequestOptions or error string.
   */
  public function getRequestOptions(string $user_email): string {
    try {
      $allowed_credentials = $this->getAllowedCredentials($user_email);
      $public_key_credential_request_options = PublicKeyCredentialRequestOptions::create(
        $this->createChallenge(),
        allowCredentials: $allowed_credentials,
      );
      $this->tempstore->set(self::TEMP_STORE_REQUEST_OPTIONS, $public_key_credential_request_options);

      $stringified_json = json_encode($public_key_credential_request_options);
      if (is_string($stringified_json)) {
        return $stringified_json;
      }
      throw new DecoupledPasskeysUnexpectedValueException('Failed to encode JSON!');
    }
    catch (DecoupledPasskeysException $e) {
      if ($e instanceof DecoupledPasskeysInvalidArgumentException) {
        return $e->getMessage();
      }
      throw new DecoupledPasskeysException('Unexpected exception getting request!', 0, $e);
    }
  }

  /**
   * Creates a Drupal public_key_credential_source entity for the given input.
   *
   * @param \Webauthn\PublicKeyCredentialSource $public_key_credential_source
   *   The public key credential source sent via JSON:RPC.
   */
  private function createDrupalEntityPublicKeyCredentialSource(PKCredentialSource $public_key_credential_source): void {
    $public_key_credential_source_entity = PublicKeyCredentialSource::createFromSourceObject($public_key_credential_source);
    $storage = $this->entityTypeManager->getStorage('user');
    $user_array = $storage->loadByProperties(['uuid' => $public_key_credential_source->userHandle]);
    $user = reset($user_array);

    if (!($user instanceof UserInterface)) {
      $key = $public_key_credential_source->publicKeyCredentialId;
      $handle = $public_key_credential_source->userHandle;
      throw new DecoupledPasskeysInvalidArgumentException("Cannot save credential source $key, the user $handle does not exist.");
    }

    $public_key_credential_source_entity->save();
  }

  /**
   * Registers a device to a user account.
   *
   * @param \Drupal\user\UserInterface $user
   *   The user to register a device for.
   * @param string $response
   *   The attestation response.
   */
  public function registerDevice(UserInterface $user, string $response): void {
    $public_key_credential = $this->publicKeyCredentialLoader->load($response);

    if ($public_key_credential->response instanceof AuthenticatorAttestationResponse) {
      /** @var \Webauthn\PublicKeyCredentialCreationOptions $saved_credentials_for_validation */
      $saved_credentials_for_validation = $this->tempstore->get(self::TEMP_STORE_CREATION_OPTIONS);

      // This will throw an exception if something is wrong.
      $public_key_credential_source = $this->authenticatorAttestationResponseValidator->check(
        $public_key_credential->response,
        $saved_credentials_for_validation,
        $this->getDrupalDomain(),
      );
      $uid = $user->id();
      $this->logger->debug("Valid public key credential returned for uid $uid!");
      $this->createDrupalEntityPublicKeyCredentialSource($public_key_credential_source);
    }
    else {
      throw new DecoupledPasskeysInvalidArgumentException('Not an AuthenticatorAttestationResponse!');
    }
  }

  /**
   * Finds a credential if it exists in the database.
   *
   * This used to be a method of PublicKeyCredentialSourceRepository,
   * findOneByCredentialId.
   *
   * @param \Webauthn\PublicKeyCredential $public_key_credential
   *   The public key credential ID to locate.
   *
   * @return \Drupal\public_key_credential_source\Entity\PublicKeyCredentialSource|null
   *   The public key credential source entity, if available.
   */
  private function getPublicKeyCredentialSourceFromId(PKCredential $public_key_credential): ?PublicKeyCredentialSource {
    $query = $this->publicKeyCredentialSourceEntityStorage->getQuery()->accessCheck(TRUE);
    $results = $query->condition('public_key_credential_source', $public_key_credential->id, 'CONTAINS')
      ->execute();

    // Stop processing if there are no results.
    if (!is_array($results)) {
      return NULL;
    }
    if (count($results) > 1) {
      throw new DecoupledPasskeysInvalidArgumentException('More than one public key credential source was returned!');
    }
    // Reindex the array so that the first value will be at the 0 position.
    $results = array_values($results);

    $entity = $this->publicKeyCredentialSourceEntityStorage->load($results[0]);
    /** @var \Drupal\public_key_credential_source\Entity\PublicKeyCredentialSource $entity */

    return $entity;
  }

  /**
   * Authenticates via passkey.
   *
   * You need to supply the user handle, which could be a Drupal uid,
   * email address, or uuid, depending on your site configuration.
   *
   * @param string $user_email
   *   The webauthn user handle. At present, only email address is supported.
   * @param string $request
   *   The passkey request.
   *
   * @returns PKCredentialSource|null
   *   The authenticated public key credential source.
   */
  public function authenticateRequest(string $user_email, string $request): ?PKCredentialSource {
    $public_key_credential = $this->publicKeyCredentialLoader->load($request);
    $user = $this->getUserFromEmail($user_email);
    if (!$user instanceof UserInterface) {
      // There is no user for the provided email address.
      return NULL;
    }
    // The userHandle for verification is the user uuid, which is used as the
    // webauthn user id when creating the key.
    $user_handle = $user->uuid();

    if ($public_key_credential->response instanceof AuthenticatorAssertionResponse) {
      /** @var \Webauthn\PublicKeyCredentialRequestOptions $saved_credentials_for_validation */
      $saved_credentials_for_validation = $this->tempstore->get(self::TEMP_STORE_REQUEST_OPTIONS);

      $public_key_credential_source_entity = $this->getPublicKeyCredentialSourceFromId($public_key_credential);

      if ($public_key_credential_source_entity === NULL) {
        throw new DecoupledPasskeysUnexpectedValueException('Failed to get public key credential source entity!');
      }

      $public_key_credential_source_from_drupal_entity = $public_key_credential_source_entity->getPublicKeyCredentialSource();

      try {
        $public_key_credential_source_webauthn = $this->authenticatorAssertionResponseValidator->check(
          $public_key_credential_source_from_drupal_entity,
          $public_key_credential->response,
          $saved_credentials_for_validation,
          $this->getDrupalDomain(),
          $user_handle,
        );

        $this->logger->debug("Valid public key credential returned for user_handle $user_handle!");

        // Resave the Drupal entity to update the last device used info.
        $public_key_credential_source_entity->save();

        return $public_key_credential_source_webauthn;
      }
      catch (\Exception $e) {
        $exception_message = $e->getMessage();
        throw new DecoupledPasskeysValidationFailedException("Webauthn authentication failed! $exception_message", 0, $e);
      }
    }
    else {
      throw new DecoupledPasskeysInvalidArgumentException('Not an AuthenticatorAssertionResponse!');
    }
  }

  /**
   * Gets the Drupal user id for a given email address.
   *
   * @param string $mail
   *   The mail address to get the user ID for.
   *
   * @return \Drupal\user\UserInterface|null
   *   The user account if found; null otherwise.
   */
  private function getUserFromEmail(string $mail): UserInterface|null {
    $user_storage = $this->entityTypeManager->getStorage('user');
    $users = $user_storage->loadByProperties(['mail' => $mail]);
    $user = reset($users);
    if ($user instanceof UserInterface) {
      return $user;
    }
    return NULL;
  }

}
